@@ -1,9 +1,13 @@
 #pragma once
-#define NO_XSLOCKS          // dont hook libc calls
-#define PERLIO_NOT_STDIO 0  // dont hook IO
-#define PERL_NO_GET_CONTEXT // we want efficiency for threaded perls
+#define NO_XSLOCKS          /* dont hook libc calls */
+#define PERLIO_NOT_STDIO 0  /* dont hook IO */
+#define PERL_NO_GET_CONTEXT /* we want efficiency for threaded perls */
 
 #ifdef __cplusplus
+    // perl breaks <algorithm> because of "seed" macro, so include it earlier to help XS code forget about this issue
+    // we can't just undef this macro as we do for "do_open", because "seed" is used by other macros
+    #include <algorithm>
+
     extern "C" {
 #endif
 
@@ -23,14 +27,6 @@
 #endif
 
 #ifdef __cplusplus
-    #undef do_open
+    #undef do_open  // perl's macros do_open, do_close breaks <iostream>, undef it as they are not used in general
     #undef do_close
-
-    #ifdef seed
-        #undef seed
-        #include <algorithm>
-        #define seed() Perl_seed(aTHX)
-    #else
-        #include <algorithm>
-    #endif
 #endif
